"""Module which defines classes all about geometry"""


class Point:
    """Data class which holds a coordinate."""

    def __init__(self, x, y):
        self.x = x
        self.y = y

    def __eq__(self, other):
        return (self.x, self.y) == (other.x, other.y)


class Distance:
    """Data class which holds the distance values in all directions."""

    def __init__(self, top=0, left=0, bottom=0, right=0):
        self.top = top
        self.left = left
        self.bottom = bottom
        self.right = right



"""This module reimplements the ThreadPoolExecutor.
https://github.com/python/cpython/blob/master/Lib/concurrent/futures/thread.py

The only change is the prevention of waiting
for each thread to exit on exiting the script.
"""

import threading
import weakref
import concurrent.futures as futures


def _worker(executor_reference, work_queue):
    # pylint: disable=W0212
    try:
        while True:
            work_item = work_queue.get(block=True)
            if work_item is not None:
                work_item.run()
                del work_item
                continue
            executor = executor_reference()
            if executor is None or executor._shutdown:
                if executor is not None:
                    executor._shutdown = True
                work_queue.put(None)
                return
            del executor
    except BaseException:
        futures._base.LOGGER.critical("Exception in worker", exc_info=True)


class DaemonThreadPoolExecutor(futures.ThreadPoolExecutor):
    """The concurrent.futures.ThreadPoolExecutor extended by
    the prevention of waiting for each thread on exiting the script.
    """

    def _adjust_thread_count(self):
        def weakref_cb(_, queue=self._work_queue):
            queue.put(None)

        num_threads = len(self._threads)
        if num_threads < self._max_workers:
            thread_name = "%s_%d" % (self, num_threads)
            thread = threading.Thread(
                name=thread_name,
                target=_worker,
                args=(weakref.ref(self, weakref_cb), self._work_queue),
            )
            thread.daemon = True
            thread.start()
            self._threads.add(thread)



class LazyConstant:
    def __init__(self, function):
        self.value = None
        self.function = function

    def __get__(self, instance, owner):
        if self.value is None:
            self.value = self.function()
        return self.value

    def __set__(self, instance, value):
        raise AttributeError("can't set attribute")



import os
import signal
import errno


def get_command(pid):
    """Figures out the associated command name
    of a process with the given pid.

    Args:
        pid (int): the pid of the process of interest

    Returns:
        str: the associated command name
    """
    with open("/proc/{}/comm".format(pid), "r") as commfile:
        return "\n".join(commfile.readlines())


def is_same_command(pid0, pid1):
    """Checks whether the associated command name
    of the processes of the given pids equals to each other.

    Args:
        pid0 (int): the pid of the process of interest
        pid1 (int): the pid of another process of interest

    Returns:
        bool: True if both processes have
              the same associated command name
    """
    return get_command(pid0) == get_command(pid1)


def send_signal_safe(own_pid, target_pid):
    """Sends SIGUSR1 to a process if both
    processes have the same associated command name.
    (Race condition free)

    Requires:
        - Python 3.9+
        - Linux 5.1+

    Args:
        own_pid (int): the pid of this process
        target_pid (int):
            the pid of the process to send the signal to
    """
    pidfile = None
    try:
        pidfile = os.open(f"/proc/{target_pid}", os.O_DIRECTORY)
        if is_same_command(own_pid, target_pid):
            signal.pidfd_send_signal(pidfile, signal.SIGUSR1)
    except FileNotFoundError:
        pass
    except OSError as error:
        # not sure if errno is really set..
        # at least the documentation of the used functions says so..
        # see e.g.: https://github.com/python/cpython/commit/7483451577916e693af6d20cf520b2cc7e2174d2#diff-99fb04b208835118fdca0d54b76a00c450da3eaff09d2b53e8a03d63bbe88e30R1279-R1281
        # and https://docs.python.org/3/c-api/exceptions.html#c.PyErr_SetFromErrno

        # caused by either pidfile_open or pidfd_send_signal
        if error.errno != errno.ESRCH:
            raise
        # else: the process is death
    finally:
        if pidfile is not None:
            os.close(pidfile)


def send_signal_unsafe(own_pid, target_pid):
    """Sends SIGUSR1 to a process if both
    processes have the same associated command name.
    (Race condition if process dies)

    Args:
        own_pid (int): the pid of this process
        target_pid (int):
            the pid of the process to send the signal to
    """
    try:
        if is_same_command(own_pid, target_pid):
            os.kill(target_pid, signal.SIGUSR1)
    except (FileNotFoundError, ProcessLookupError):
        pass


def main(options):
    # assumption:
    # started by calling the programs name
    # ueberzug layer and
    # ueberzug query_windows
    own_pid = os.getpid()

    for pid in options["PIDS"]:
        try:
            send_signal_safe(own_pid, int(pid))
        except AttributeError:
            send_signal_unsafe(own_pid, int(pid))



import abc
import enum
import subprocess
import threading
import json
import collections
import contextlib
import os
import signal

import attr

import ueberzug.action as _action
from ueberzug.scaling import ScalerOption
from ueberzug.loading import ImageLoaderOption


class Visibility(enum.Enum):
    """Enum which defines the different visibility states."""

    VISIBLE = enum.auto()
    INVISIBLE = enum.auto()


class Placement:
    """The class which represent a (image) placement on the canvas.

    Attributes:
        Every parameter defined by the add action is an attribute.

    Raises:
        IOError: on assign a new value to an attribute
                 if stdin of the ueberzug process was closed
                 during an attempt of writing to it.
    """

    __initialised = False
    __DEFAULT_VALUES = {str: "", int: 0}
    __ATTRIBUTES = {
        attribute.name: attribute
        for attribute in attr.fields(_action.AddImageAction)
    }
    __EMPTY_BASE_PAIRS = (
        lambda attributes, default_values: {
            attribute.name: default_values[attribute.type]
            for attribute in attributes.values()
            if (attribute.default == attr.NOTHING and attribute.init)
        }
    )(__ATTRIBUTES, __DEFAULT_VALUES)

    def __init__(
        self,
        canvas,
        identifier,
        visibility: Visibility = Visibility.INVISIBLE,
        **kwargs,
    ):
        """
        Args:
            canvas (Canvas): the canvas this placement belongs to
            identifier (str): a string which uniquely identifies this placement
            visibility (Visibility): the initial visibility of this placement
                                     (all required parameters need to be set
                                     if it's visible)
            kwargs: parameters of the add action
        """
        self.__canvas = canvas
        self.__identifier = identifier
        self.__visibility = False
        self.__data = {}
        self.__initialised = True
        for key, value in kwargs.items():
            setattr(self, key, value)
        self.visibility = visibility

    @property
    def canvas(self):
        """Canvas: the canvas this placement belongs to"""
        return self.__canvas

    @property
    def identifier(self):
        """str: the identifier of this placement"""
        return self.__identifier

    @property
    def visibility(self):
        """Visibility: the visibility of this placement"""
        return self.__visibility

    @visibility.setter
    def visibility(self, value):
        if self.__visibility != value:
            if value is Visibility.INVISIBLE:
                self.__remove()
            elif value is Visibility.VISIBLE:
                self.__update()
            else:
                raise TypeError("expected an instance of Visibility")
            self.__visibility = value

    def __remove(self):
        self.__canvas.enqueue(
            _action.RemoveImageAction(identifier=self.identifier)
        )
        self.__canvas.request_transmission()

    def __update(self):
        self.__canvas.enqueue(
            _action.AddImageAction(
                **{
                    **self.__data,
                    **attr.asdict(
                        _action.Identifiable(identifier=self.identifier)
                    ),
                }
            )
        )
        self.__canvas.request_transmission()

    def __getattr__(self, name):
        if name not in self.__ATTRIBUTES:
            raise AttributeError("There is no attribute named %s" % name)

        attribute = self.__ATTRIBUTES[name]

        if name in self.__data:
            return self.__data[name]
        if attribute.default != attr.NOTHING:
            return attribute.default
        return None

    def __setattr__(self, name, value):
        if not self.__initialised:
            super().__setattr__(name, value)
            return
        if name not in self.__ATTRIBUTES:
            if hasattr(self, name):
                super().__setattr__(name, value)
                return
            raise AttributeError("There is no attribute named %s" % name)

        data = dict(self.__data)
        self.__data.update(
            attr.asdict(
                _action.AddImageAction(
                    **{
                        **self.__EMPTY_BASE_PAIRS,
                        **self.__data,
                        **attr.asdict(
                            _action.Identifiable(identifier=self.identifier)
                        ),
                        name: value,
                    }
                )
            )
        )

        # remove the key's of the empty base pairs
        # so the developer is forced to set them by himself
        for key in self.__EMPTY_BASE_PAIRS:
            if key not in data and key != name:
                del self.__data[key]

        if self.visibility is Visibility.VISIBLE:
            self.__update()


class UeberzugProcess:
    """Class which handles the creation and
    destructions of ueberzug processes.
    """

    __KILL_TIMEOUT_SECONDS = 1
    __BUFFER_SIZE_BYTES = 50 * 1024

    def __init__(self, options):
        """
        Args:
            options (list of str): additional command line arguments
        """
        self.__start_options = options
        self.__process = None

    @property
    def stdin(self):
        """_io.TextIOWrapper: stdin of the ueberzug process"""
        return self.__process.stdin

    @property
    def running(self):
        """bool: ueberzug process is still running"""
        return self.__process is not None and self.__process.poll() is None

    @property
    def responsive(self):
        """bool: ueberzug process is able to receive instructions"""
        return self.running and not self.__process.stdin.closed

    def start(self):
        """Starts a new ueberzug process
        if there's none or it's not responsive.
        """
        if self.responsive:
            return
        if self.running:
            self.stop()

        self.__process = subprocess.Popen(
            ["ueberzug", "layer"] + self.__start_options,
            stdin=subprocess.PIPE,
            bufsize=self.__BUFFER_SIZE_BYTES,
            universal_newlines=True,
            start_new_session=True,
        )

    def stop(self):
        """Sends SIGTERM to the running ueberzug process
        and waits for it to exit.
        If the process won't end after a specific timeout
        SIGKILL will also be send.
        """
        if self.running:
            timer_kill = None

            try:
                ueberzug_pgid = os.getpgid(self.__process.pid)
                own_pgid = os.getpgid(0)
                assert ueberzug_pgid != own_pgid
                timer_kill = threading.Timer(
                    self.__KILL_TIMEOUT_SECONDS,
                    os.killpg,
                    [ueberzug_pgid, signal.SIGKILL],
                )

                self.__process.terminate()
                timer_kill.start()
                self.__process.communicate()
            except ProcessLookupError:
                pass
            finally:
                if timer_kill is not None:
                    timer_kill.cancel()


class CommandTransmitter:
    """Describes the structure used to define command transmitter classes.

    Defines a general interface used to implement different ways
    of storing and transmitting commands to ueberzug processes.
    """

    def __init__(self, process):
        self._process = process

    @abc.abstractproperty
    def synchronously_draw(self):
        """bool: execute draw operations of ImageActions synchrously"""
        raise NotImplementedError()

    @abc.abstractmethod
    def enqueue(self, action: _action.Action):
        """Enqueues a command.

        Args:
            action (action.Action): the command which should be executed
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def transmit(self):
        """Transmits every command in the queue."""
        raise NotImplementedError()


class DequeCommandTransmitter(CommandTransmitter):
    """Implements the command transmitter with a dequeue."""

    def __init__(self, process):
        super().__init__(process)
        self.__queue_commands = collections.deque()
        self.__synchronously_draw = False

    @property
    def synchronously_draw(self):
        return self.__synchronously_draw

    @synchronously_draw.setter
    def synchronously_draw(self, value):
        self.__synchronously_draw = value

    def enqueue(self, action: _action.Action):
        self.__queue_commands.append(action)

    def transmit(self):
        while self.__queue_commands:
            command = self.__queue_commands.popleft()
            self._process.stdin.write(
                json.dumps(
                    {
                        **attr.asdict(command),
                        **attr.asdict(
                            _action.Drawable(
                                synchronously_draw=self.__synchronously_draw,
                                draw=not self.__queue_commands,
                            )
                        ),
                    }
                )
            )
            self._process.stdin.write("\n")
        self._process.stdin.flush()


class LazyCommandTransmitter(CommandTransmitter):
    """Implements lazily transmitting commands as decorator class.

    Ignores calls of the transmit method.
    """

    def __init__(self, transmitter):
        super().__init__(None)
        self.transmitter = transmitter

    @property
    def synchronously_draw(self):
        return self.transmitter.synchronously_draw

    @synchronously_draw.setter
    def synchronously_draw(self, value):
        self.transmitter.synchronously_draw = value

    def enqueue(self, action: _action.Action):
        self.transmitter.enqueue(action)

    def transmit(self):
        pass

    def force_transmit(self):
        """Executes the transmit method of the decorated CommandTransmitter."""
        self.transmitter.transmit()


class Canvas:
    """The class which represents the drawing area."""

    def __init__(self, debug=False):
        self.__process_arguments = (
            ["--loader", ImageLoaderOption.SYNCHRONOUS.value]
            if debug
            else ["--silent"]
        )
        self.__process = None
        self.__transmitter = None
        self.__used_identifiers = set()
        self.automatic_transmission = True

    def create_placement(self, identifier, *args, **kwargs):
        """Creates a placement associated with this canvas.

        Args:
            the same as the constructor of Placement
        """
        if identifier in self.__used_identifiers:
            raise ValueError("Identifier '%s' is already taken." % identifier)
        self.__used_identifiers.add(identifier)
        return Placement(self, identifier, *args, **kwargs)

    @property
    @contextlib.contextmanager
    def lazy_drawing(self):
        """Context manager factory function which
        prevents transmitting commands till the with-statement ends.

        Raises:
            IOError: on transmitting commands
                     if stdin of the ueberzug process was closed
                     during an attempt of writing to it.
        """
        try:
            self.__transmitter.transmit()
            self.__transmitter = LazyCommandTransmitter(self.__transmitter)
            yield
            self.__transmitter.force_transmit()
        finally:
            self.__transmitter = self.__transmitter.transmitter

    @property
    @contextlib.contextmanager
    def synchronous_lazy_drawing(self):
        """Context manager factory function which
        prevents transmitting commands till the with-statement ends.
        Also enforces to execute the draw operation synchronously
        right after the last command.

        Raises:
            IOError: on transmitting commands
                     if stdin of the ueberzug process was closed
                     during an attempt of writing to it.
        """
        try:
            self.__transmitter.synchronously_draw = True
            with self.lazy_drawing:
                yield
        finally:
            self.__transmitter.synchronously_draw = False

    def __call__(self, function):
        def decorator(*args, **kwargs):
            with self:
                return function(*args, canvas=self, **kwargs)

        return decorator

    def __enter__(self):
        self.__process = UeberzugProcess(self.__process_arguments)
        self.__transmitter = DequeCommandTransmitter(self.__process)
        self.__process.start()
        return self

    def __exit__(self, *args):
        try:
            self.__process.stop()
        finally:
            self.__process = None
            self.__transmitter = None

    def enqueue(self, command: _action.Action):
        """Enqueues a command.

        Args:
            action (action.Action): the command which should be executed
        """
        if not self.__process.responsive:
            self.__process.start()

        self.__transmitter.enqueue(command)

    def request_transmission(self, *, force=False):
        """Requests the transmission of every command in the queue."""
        if not self.__process.responsive:
            self.__process.start()

        if self.automatic_transmission or force:
            self.__transmitter.transmit()



import re
import os
import functools


MAX_PROCESS_NAME_LENGTH = 15


@functools.wraps(os.getpid)
def get_own_pid(*args, **kwargs):
    # pylint: disable=missing-docstring
    return os.getpid(*args, **kwargs)


def get_info(pid: int):
    """Determines information about the process with the given pid.

    Determines
    - the process id (pid)
    - the command name (comm)
    - the state (state)
    - the process id of the parent process (ppid)
    - the process group id (pgrp)
    - the session id (session)
    - the controlling terminal (tty_nr)
    of the process with the given pid.

    Args:
        pid (int or str):
            the associated pid of the process
            for which to retrieve the information for

    Returns:
        dict of str: bytes:
            containing the listed information.
            The term in the brackets is used as key.

    Raises:
        FileNotFoundError: if there is no process with the given pid
    """
    with open(f"/proc/{pid}/stat", "rb") as proc_file:
        data = proc_file.read()
        return re.search(
            rb"^(?P<pid>[-+]?\d+) "
            rb"\((?P<comm>.{0,"
            + str(MAX_PROCESS_NAME_LENGTH).encode()
            + rb"})\) "
            rb"(?P<state>.) "
            rb"(?P<ppid>[-+]?\d+) "
            rb"(?P<pgrp>[-+]?\d+) "
            rb"(?P<session>[-+]?\d+) "
            rb"(?P<tty_nr>[-+]?\d+)",
            data,
            re.DOTALL,
        ).groupdict()


@functools.lru_cache()
def get_pty_slave_folders():
    """Determines the folders in which linux
    creates the control device files of the pty slaves.

    Returns:
        list of str: containing the paths to these folders
    """
    paths = []

    with open("/proc/tty/drivers", "rb") as drivers_file:
        for line in drivers_file:
            # The documentation about /proc/tty/drivers
            # is a little bit short (man proc):
            # /proc/tty
            #     Subdirectory containing the pseudo-files and
            #     subdirectories for tty drivers and line disciplines.
            # So.. see the source code:
            # https://github.com/torvalds/linux/blob/8653b778e454a7708847aeafe689bce07aeeb94e/fs/proc/proc_tty.c#L28-L67
            driver = re.search(
                rb"^(?P<name>(\S| )+?) +" rb"(?P<path>/dev/\S+) ", line
            ).groupdict()
            if driver["name"] == b"pty_slave":
                paths += [driver["path"].decode()]

    return paths


def get_parent_pid(pid: int):
    """Determines pid of the parent process of the process with the given pid.

    Args:
        pid (int or str):
            the associated pid of the process
            for which to retrieve the information for

    Returns:
        int: the pid of the parent process

    Raises:
        FileNotFoundError: if there is no process with the given pid
    """
    process_info = get_info(pid)
    return int(process_info["ppid"])


def calculate_minor_device_number(tty_nr: int):
    """Calculates the minor device number contained
    in a tty_nr of a stat file of the procfs.

    Args:
        tty_nr (int):
            a tty_nr of a stat file of the procfs

    Returns:
        int: the minor device number contained in the tty_nr
    """
    TTY_NR_BITS = 32
    FIRST_BITS = 8
    SHIFTED_BITS = 12
    FIRST_BITS_MASK = 0xFF
    SHIFTED_BITS_MASK = 0xFFF00000
    minor_device_number = (tty_nr & FIRST_BITS_MASK) + (
        (tty_nr & SHIFTED_BITS_MASK)
        >> (TTY_NR_BITS - SHIFTED_BITS - FIRST_BITS)
    )
    return minor_device_number


def get_pty_slave(pid: int):
    """Determines control device file
    of the pty slave of the process with the given pid.

    Args:
        pid (int or str):
            the associated pid of the process
            for which to retrieve the information for

    Returns:
        str or None:
            the path to the control device file
            or None if no path was found

    Raises:
        FileNotFoundError: if there is no process with the given pid
    """
    pty_slave_folders = get_pty_slave_folders()
    process_info = get_info(pid)
    tty_nr = int(process_info["tty_nr"])
    minor_device_number = calculate_minor_device_number(tty_nr)

    for folder in pty_slave_folders:
        device_path = f"{folder}/{minor_device_number}"

        try:
            if tty_nr == os.stat(device_path).st_rdev:
                return device_path
        except FileNotFoundError:
            pass

    return None



#!/usr/bin/env python3
"""Usage:
    ueberzug layer [options]
    ueberzug library
    ueberzug version
    ueberzug query_windows PIDS ...

Routines:
    layer                   Display images
    library                 Prints the path to the bash library
    version                 Prints the project version
    query_windows           Orders ueberzug to search for windows.
                            Only for internal use.

Layer options:
    -p, --parser <parser>  one of json, simple, bash
                           json: Json-Object per line
                           simple: Key-Values separated by a tab
                           bash: associative array dumped via `declare -p`
                           [default: json]
    -l, --loader <loader>  one of synchronous, thread, process
                           synchronous: load images right away
                           thread: load images in threads
                           process: load images in additional processes
                           [default: thread]
    -s, --silent           print stderr to /dev/null


License:
    ueberzug
    This program comes with ABSOLUTELY NO WARRANTY.
    This is free software, and you are welcome to redistribute it
    under certain conditions.
"""
import docopt


def main():
    options = docopt.docopt(__doc__)
    module = None

    if options["layer"]:
        import ueberzug.layer as layer

        module = layer
    elif options["library"]:
        import ueberzug.library as library

        module = library
    elif options["query_windows"]:
        import ueberzug.query_windows as query_windows

        module = query_windows
    elif options["version"]:
        import ueberzug.version as version

        module = version

    module.main(options)


if __name__ == "__main__":
    main()



# extracted from distutils
# under license: MIT
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.
#
def strtobool(val):
    """Convert a string representation of truth to true (1) or false (0).

    True values are 'y', 'yes', 't', 'true', 'on', and '1'; false values
    are 'n', 'no', 'f', 'false', 'off', and '0'.  Raises ValueError if
    'val' is anything else.
    """
    val = val.lower()
    if val in ("y", "yes", "t", "true", "on", "1"):
        return 1
    elif val in ("n", "no", "f", "false", "off", "0"):
        return 0
    else:
        raise ValueError(f"invalid truth value {val!r}")


def to_bool(value):
    """Converts a String to a Boolean.

    Args:
        value (str or bool): a boolean or a string representing a boolean

    Returns:
        bool: the evaluated boolean
    """
    return value if isinstance(value, bool) else bool(strtobool(value))
